@@ -110,7 +110,7 @@ ch_util_get_descriptions (GPtrArray *array)
 	GString *string;
 
 	/* get maximum command length */
-	for (i=0; i<array->len; i++) {
+	for (i = 0; i < array->len; i++) {
 		item = g_ptr_array_index (array, i);
 		len = strlen (item->name);
 		if (len > max_len)
@@ -123,7 +123,7 @@ ch_util_get_descriptions (GPtrArray *array)
 
 	/* print each command */
 	string = g_string_new ("");
-	for (i=0; i<array->len; i++) {
+	for (i = 0; i < array->len; i++) {
 		item = g_ptr_array_index (array, i);
 		g_string_append (string, "  ");
 		g_string_append (string, item->name);
@@ -153,7 +153,7 @@ ch_util_run (ChUtilPrivate *priv, const gchar *command, gchar **values, GError *
 	GString *string;
 
 	/* find command */
-	for (i=0; i<priv->cmd_array->len; i++) {
+	for (i = 0; i < priv->cmd_array->len; i++) {
 		item = g_ptr_array_index (priv->cmd_array, i);
 		if (g_strcmp0 (item->name, command) == 0) {
 			ret = item->callback (priv, values, error);
@@ -166,7 +166,7 @@ ch_util_run (ChUtilPrivate *priv, const gchar *command, gchar **values, GError *
 	/* TRANSLATORS: error message */
 	g_string_append_printf (string, "%s\n",
 				_("Command not found, valid commands are:"));
-	for (i=0; i<priv->cmd_array->len; i++) {
+	for (i = 0; i < priv->cmd_array->len; i++) {
 		item = g_ptr_array_index (priv->cmd_array, i);
 		g_string_append_printf (string, " * %s\n", item->name);
 	}
@@ -487,7 +487,7 @@ ch_util_remote_profile_upload (ChUtilPrivate *priv, gchar **values, GError **err
 			     msg->reason_phrase);
 		goto out;
 	}
-	uri = soup_message_headers_get (msg->response_headers, "Location");
+	uri = soup_message_headers_get_one (msg->response_headers, "Location");
 	g_debug ("Successfully uploaded to %s", uri);
 
 	/* print something machine readable */
@@ -528,6 +528,70 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_ccmx_upload:
+ **/
+static gboolean
+ch_util_ccmx_upload (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	const gchar *uri;
+	gboolean ret = TRUE;
+	gchar *data = NULL;
+	gsize length;
+	guint status_code;
+	SoupBuffer *buffer = NULL;
+	SoupMessage *msg = NULL;
+	SoupMultipart *multipart = NULL;
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'filename.ccmx'");
+		goto out;
+	}
+
+	/* read file */
+	ret = g_file_get_contents (values[0], &data, &length, error);
+	if (!ret)
+		goto out;
+
+	/* create multipart form and upload file */
+	multipart = soup_multipart_new (SOUP_FORM_MIME_TYPE_MULTIPART);
+	buffer = soup_buffer_new (SOUP_MEMORY_STATIC, data, length);
+	soup_multipart_append_form_file (multipart,
+					 "upload",
+					 values[0],
+					 NULL,
+					 buffer);
+	msg = soup_form_request_new_from_multipart ("http://www.hughski.com/ccmx-store.php", multipart);
+	status_code = soup_session_send_message (priv->session, msg);
+	if (!SOUP_STATUS_IS_SUCCESSFUL (status_code)) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "Failed to upload file: %s",
+			     msg->reason_phrase);
+		goto out;
+	}
+	uri = soup_message_headers_get_one (msg->response_headers, "Location");
+	g_debug ("Successfully uploaded to %s", uri);
+
+	/* print something machine readable */
+	if (g_strcmp0 (g_getenv ("COLORHUG_OUTPUT"), "plain") == 0)
+		g_print ("%s\n", uri);
+	else
+		g_print ("Uploaded CCMX to %s\n", uri);
+out:
+	if (buffer != NULL)
+		soup_buffer_free (buffer);
+	if (msg != NULL)
+		g_object_unref (msg);
+	if (multipart != NULL)
+		soup_multipart_free (multipart);
+	g_free (data);
+	return ret;
+}
+
 /**
  * ch_util_set_color_select:
  **/
@@ -1343,8 +1407,8 @@ out:
 static gboolean
 ch_util_set_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 {
-	gboolean ret;
-	guint8 leds;
+	gboolean ret = FALSE;
+	guint8 leds = 0;
 	guint8 repeat = 0;
 	guint8 time_on = 0x00;
 	guint8 time_off = 0x00;
@@ -1352,30 +1416,39 @@ ch_util_set_leds (ChUtilPrivate *priv, gchar **values, GError **error)
 	/* parse */
 	if (g_strv_length (values) != 1 &&
 	    g_strv_length (values) != 4) {
-		ret = FALSE;
 		g_set_error_literal (error, 1, 0,
 				     "invalid input, expect "
-				     "'<red|green|both> <repeat> <time_on> <time_off>' or "
+				     "'<red|green|blue|white|off> <repeat> <time_on> <time_off>' or "
 				     "'<leds>'");
 		goto out;
 	}
 
 	/* get the LEDs value */
-	if (g_strcmp0 (values[0], "red") == 0) {
-		leds = CH_STATUS_LED_RED;
-	} else if (g_strcmp0 (values[0], "green") == 0) {
-		leds = CH_STATUS_LED_GREEN;
-	} else if (g_strcmp0 (values[0], "both") == 0) {
-		leds = CH_STATUS_LED_RED | CH_STATUS_LED_GREEN;
-	} else {
-		leds = g_ascii_strtoull (values[0], NULL, 10);
-		if (leds > 3) {
-			ret = FALSE;
-			g_set_error (error, 1, 0,
-				     "invalid leds value %i",
-				     leds);
-			goto out;
-		}
+	if (g_strstr_len (values[0], -1, "red") != NULL) {
+		leds |= CH_STATUS_LED_RED;
+		ret = TRUE;
+	}
+	if (g_strstr_len (values[0], -1, "green") != NULL) {
+		leds |= CH_STATUS_LED_GREEN;
+		ret = TRUE;
+	}
+	if (g_strstr_len (values[0], -1, "blue") != NULL) {
+		leds |= CH_STATUS_LED_BLUE;
+		ret = TRUE;
+	}
+	if (g_strstr_len (values[0], -1, "white") != NULL) {
+		leds |= CH_STATUS_LED_RED |
+			CH_STATUS_LED_GREEN |
+			CH_STATUS_LED_BLUE;
+		ret = TRUE;
+	}
+
+	/* nothing recognised */
+	if (!ret) {
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect "
+				     "'<red|green|blue|white|off>");
+		goto out;
 	}
 
 	/* get the optional other parameters */
@@ -1593,7 +1666,7 @@ ch_util_set_dark_offsets_auto (ChUtilPrivate *priv, GError **error)
 	CdColorRGB value_zero;
 	CdColorRGB value;
 
-	/* TRANSLATORS: wait for user to press the device into a desk... */
+	/* TRANSLATORS: wait for user to press the device into a desk… */
 	ret = ch_util_get_prompt (_("Ensure the ColorHug aperture is blocked"), TRUE);
 	if (!ret) {
 		g_set_error_literal (error, 1, 0,
@@ -1602,7 +1675,7 @@ ch_util_set_dark_offsets_auto (ChUtilPrivate *priv, GError **error)
 	}
 
 	/* set dark offsets */
-	cd_color_set_rgb (&value_zero, 0.0f, 0.0f, 0.0f);
+	cd_color_rgb_set (&value_zero, 0.0f, 0.0f, 0.0f);
 
 	/* get from HW */
 	ch_device_queue_get_dark_offsets (priv->device_queue,
@@ -1757,9 +1830,10 @@ ch_util_take_reading_raw (ChUtilPrivate *priv, gchar **values, GError **error)
 {
 	ChColorSelect color_select = 0;
 	ChFreqScale multiplier = 0;
+	ChMeasureMode measure_mode = 0;
 	gboolean ret;
 	guint16 integral_time = 0;
-	guint16 take_reading;
+	guint32 take_reading;
 
 	/* get from HW */
 	ch_device_queue_get_color_select (priv->device_queue,
@@ -1768,6 +1842,9 @@ ch_util_take_reading_raw (ChUtilPrivate *priv, gchar **values, GError **error)
 	ch_device_queue_get_multiplier (priv->device_queue,
 					priv->device,
 					&multiplier);
+	ch_device_queue_get_measure_mode (priv->device_queue,
+					  priv->device,
+					  &measure_mode);
 	ch_device_queue_get_integral_time (priv->device_queue,
 					   priv->device,
 					   &integral_time);
@@ -1775,7 +1852,7 @@ ch_util_take_reading_raw (ChUtilPrivate *priv, gchar **values, GError **error)
 					  priv->device,
 					  &take_reading);
 	ret = ch_device_queue_process (priv->device_queue,
-				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_CONTINUE_ERRORS,
 				       NULL,
 				       error);
 	if (!ret)
@@ -1789,11 +1866,15 @@ ch_util_take_reading_raw (ChUtilPrivate *priv, gchar **values, GError **error)
 	g_print ("%s:\t%s\n", _("Multiplier"),
 		 ch_multiplier_to_string (multiplier));
 
+	/* TRANSLATORS: this is the measurement mode */
+	g_print ("%s:\t%s\n", _("Measure mode"),
+		 ch_measure_mode_to_string (measure_mode));
+
 	/* TRANSLATORS: this is the sensor sample time */
 	g_print ("%s:\t0x%04x\n", _("Integral"), integral_time);
 
 	/* TRANSLATORS: this is the number of pulses detected */
-	g_print ("%s:\t\t%i\n", _("Pulses"), take_reading);
+	g_print ("%s:\t\t%" G_GUINT32_FORMAT "\n", _("Pulses"), take_reading);
 out:
 	return ret;
 }
@@ -1884,6 +1965,68 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_take_reading_spectral:
+ **/
+static gboolean
+ch_util_take_reading_spectral (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gdouble scale;
+	gsize len;
+	guint16 *data = NULL;
+	guint16 largest = 0;
+	guint16 sram_addr = 0x0000;
+	guint i;
+	guint j;
+
+	/* get from HW */
+	ch_device_queue_set_integral_time (priv->device_queue,
+					    priv->device,
+					    50);
+	ch_device_queue_take_reading_spectral (priv->device_queue,
+						 priv->device,
+						 &sram_addr);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	len = CH_CCD_SPECTRAL_RESOLUTION;
+	data = g_new0 (guint16, len);
+	g_print ("Got data at %04x\n", sram_addr);
+	ch_device_queue_read_sram (priv->device_queue,
+				   priv->device,
+				   sram_addr,
+				   (guint8 *) data,
+				   len);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	/* find biggest value */
+	for (i = 0; i < len; i++) {
+		if (data[i] > largest)
+			largest = data[i];
+	}
+	scale = 72.0f / (gdouble) largest;
+
+	g_print ("Read:\n");
+	for (i = 0; i < len; i++) {
+		for (j = 0; j < data[i] * scale; j++)
+			g_print ("#");
+		g_print ("\n");
+	}
+out:
+	g_free (data);
+	return ret;
+}
+
 /**
  * ch_util_reset:
  **/
@@ -1985,7 +2128,8 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 				 GError **error)
 {
 	gboolean ret;
-	gchar *data = NULL;
+	gchar *data_raw = NULL;
+	guint8 *data = NULL;
 	gsize len = 0;
 	GUsbDevice *device = NULL;
 	GMainLoop *loop = NULL;
@@ -1994,15 +2138,29 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
 	/* load file */
-	ret = g_file_get_contents (filename, &data, &len, error);
+	ret = g_file_get_contents (filename, &data_raw, &len, error);
 	if (!ret)
 		goto out;
 
-	/* boot to bootloader */
-	device = ch_util_get_default_device (error);
-	if (!ret)
+	/* unpack the hex file to a binary blob if required */
+	if (g_str_has_suffix (filename, ".bin")) {
+		data = g_memdup (data_raw, len);
+	} else if (g_str_has_suffix (filename, ".hex")) {
+		ret = ch_inhx32_to_bin (data_raw,
+					&data,
+					&len,
+					error);
+		if (!ret)
+			goto out;
+	} else {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid file type, expect .bin or .hex");
 		goto out;
-	ch_device_queue_reset (priv->device_queue, device);
+	}
+
+	/* boot to bootloader */
+	ch_device_queue_reset (priv->device_queue, priv->device);
 	ret = ch_device_queue_process (priv->device_queue,
 				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
 				       NULL,
@@ -2016,7 +2174,6 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 		       ch_util_helper_quit_loop_cb,
 		       loop);
 	g_main_loop_run (loop);
-	g_object_unref (device);
 	device = ch_util_get_default_device (error);
 	if (device == NULL)
 		goto out;
@@ -2027,7 +2184,7 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 					   0x00);
 	ch_device_queue_write_firmware (priv->device_queue,
 					device,
-					(const guint8 *) data,
+					data,
 					len);
 	ret = ch_device_queue_process (priv->device_queue,
 				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
@@ -2039,7 +2196,7 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 	/* read firmware */
 	ch_device_queue_verify_firmware (priv->device_queue,
 					 device,
-					 (const guint8 *) data,
+					 data,
 					 len);
 	ch_device_queue_boot_flash (priv->device_queue,
 				    device);
@@ -2057,8 +2214,10 @@ ch_util_flash_firmware_internal (ChUtilPrivate *priv,
 		       loop);
 	g_main_loop_run (loop);
 	device = ch_util_get_default_device (error);
-	if (device == NULL)
+	if (device == NULL) {
+		ret = FALSE;
 		goto out;
+	}
 
 	/* set flash success true */
 	ch_device_queue_set_flash_success (priv->device_queue,
@@ -2076,6 +2235,7 @@ out:
 	if (device != NULL)
 		g_object_unref (device);
 	g_free (data);
+	g_free (data_raw);
 	return ret;
 }
 
@@ -2213,6 +2373,141 @@ out:
 	return ret;
 }
 
+/**
+ * ch_util_get_adc_vrefs:
+ **/
+static gboolean
+ch_util_get_adc_vrefs (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gdouble vref_neg = 0.0f;
+	gdouble vref_pos = 0.0f;
+
+	/* get from HW */
+	ch_device_queue_get_adc_vref_neg (priv->device_queue,
+					  priv->device,
+					  &vref_neg);
+	ch_device_queue_get_adc_vref_pos (priv->device_queue,
+					  priv->device,
+					  &vref_pos);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	g_print ("ADC Vref+: %f Volts\n", vref_pos);
+	g_print ("ADC Vref-: %f Volts\n", vref_neg);
+out:
+	return ret;
+}
+
+/**
+ * ch_util_get_ccd_calibration:
+ **/
+static gboolean
+ch_util_get_ccd_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	guint16 ccd_calibration[3] = { 0x0000, 0x0000, 0x0000 };
+
+	/* get from HW */
+	ch_device_queue_get_ccd_calibration (priv->device_queue,
+					     priv->device,
+					     ccd_calibration);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	g_print ("CCD Calibration: %i %i %i\n",
+		 ccd_calibration[0],
+		 ccd_calibration[1],
+		 ccd_calibration[2]);
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_ccd_calibration:
+ **/
+static gboolean
+ch_util_set_ccd_calibration (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	guint16 ccd_calibration[3] = { 0x0000, 0x0000, 0x0000 };
+	guint16 last = 0;
+	guint i;
+
+	/* parse */
+	if (g_strv_length (values) != 3) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'red', 'green', 'blue'");
+		goto out;
+	}
+	for (i = 0; i < 3; i++) {
+		ccd_calibration[i] = g_ascii_strtoll (values[i], NULL, 10);
+		if (ccd_calibration[i] == 0 ||
+		    ccd_calibration[i] > CH_CCD_SPECTRAL_RESOLUTION) {
+			ret = FALSE;
+			g_set_error (error, 1, 0,
+				     "invalid ccd calibration value %s",
+				     values[i]);
+			goto out;
+		}
+		if (ccd_calibration[i] <= last) {
+			ret = FALSE;
+			g_set_error (error, 1, 0,
+				     "ccd calibration values should increase %s",
+				     values[i]);
+			goto out;
+		}
+		last = ccd_calibration[i];
+	}
+
+	/* get from HW */
+	ch_device_queue_set_ccd_calibration (priv->device_queue,
+					     priv->device,
+					     ccd_calibration);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+out:
+	return ret;
+}
+
+/**
+ * ch_util_get_temperature:
+ **/
+static gboolean
+ch_util_get_temperature (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gdouble temperature = 0.0f;
+
+	/* get from HW */
+	ch_device_queue_get_temperature (priv->device_queue,
+				         priv->device,
+				         &temperature);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	g_print ("Temperature: %f\n", temperature);
+out:
+	return ret;
+}
+
 /**
  * ch_util_get_post_scale:
  **/
@@ -2406,7 +2701,7 @@ ch_util_eeprom_write (ChUtilPrivate *priv, gchar **values, GError **error)
 		goto out;
 
 	g_print ("Wrote:\n");
-	for (i=0; i< len; i++)
+	for (i = 0; i < len; i++)
 		g_print ("0x%04x = %02x\n", address + i, data[i]);
 out:
 	g_free (data);
@@ -2514,7 +2809,204 @@ ch_util_eeprom_read (ChUtilPrivate *priv, gchar **values, GError **error)
 		goto out;
 
 	g_print ("Read:\n");
-	for (i=0; i< len; i++)
+	for (i = 0; i < len; i++)
+		g_print ("0x%04x = %02x\n", address + i, data[i]);
+out:
+	g_free (data);
+	return ret;
+}
+
+/**
+ * ch_util_get_measure_mode:
+ **/
+static gboolean
+ch_util_get_measure_mode (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	ChMeasureMode measure_mode = CH_MEASURE_MODE_FREQUENCY;
+
+	/* get from HW */
+	ch_device_queue_get_measure_mode (priv->device_queue,
+					  priv->device,
+					  &measure_mode);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	switch (measure_mode) {
+	case CH_MEASURE_MODE_FREQUENCY:
+	case CH_MEASURE_MODE_DURATION:
+		g_print ("%s\n", ch_measure_mode_to_string (measure_mode));
+		break;
+	default:
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid measure_mode value %i",
+			     measure_mode);
+	}
+out:
+	return ret;
+}
+
+/**
+ * ch_util_set_measure_mode:
+ **/
+static gboolean
+ch_util_set_measure_mode (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	ChMeasureMode measure_mode;
+
+	/* parse */
+	if (g_strv_length (values) != 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'frequency|duration'");
+		goto out;
+	}
+	if (g_strcmp0 (values[0], "frequency") == 0)
+		measure_mode = CH_MEASURE_MODE_FREQUENCY;
+	else if (g_strcmp0 (values[0], "duration") == 0)
+		measure_mode = CH_MEASURE_MODE_DURATION;
+	else {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid input '%s', expect 'frequency|duration'",
+			     values[0]);
+		goto out;
+	}
+
+	/* set to HW */
+	ch_device_queue_set_measure_mode (priv->device_queue,
+					  priv->device,
+					  measure_mode);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+out:
+	return ret;
+}
+
+/**
+ * ch_util_sram_write:
+ **/
+static gboolean
+ch_util_sram_write (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gsize len;
+	guint32 address;
+	guint8 *data = NULL;
+	guint i;
+
+	/* parse */
+	if (g_strv_length (values) != 2) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'address (base-16)' 'length (base-10)'");
+		goto out;
+	}
+
+	/* read sram */
+	address = g_ascii_strtoull (values[0], NULL, 16);
+	if (address > 0xffff) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid address 0x%04x",
+			     address);
+		goto out;
+	}
+	len = g_ascii_strtoull (values[1], NULL, 10);
+	if (len < 1) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid length");
+		goto out;
+	}
+
+	/* just write zeros */
+	data = g_new0 (guint8, len);
+	for (i = 0; i < len; i++)
+		data[i] = g_random_int_range (0x00, 0xff);
+	ch_device_queue_write_sram (priv->device_queue,
+				    priv->device,
+				    (guint16) address,
+				    data,
+				    len);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	g_print ("Wrote:\n");
+	for (i = 0; i < len; i++)
+		g_print ("0x%04x = %02x\n", address + i, data[i]);
+out:
+	g_free (data);
+	return ret;
+}
+
+/**
+ * ch_util_sram_read:
+ **/
+static gboolean
+ch_util_sram_read (ChUtilPrivate *priv, gchar **values, GError **error)
+{
+	gboolean ret;
+	gsize len;
+	guint32 address;
+	guint8 *data = NULL;
+	guint i;
+
+	/* parse */
+	if (g_strv_length (values) != 2) {
+		ret = FALSE;
+		g_set_error_literal (error, 1, 0,
+				     "invalid input, expect 'address (base-16)' 'length (base-10)'");
+		goto out;
+	}
+
+	/* read sram */
+	address = g_ascii_strtoull (values[0], NULL, 16);
+	if (address > 0xffff) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid address 0x%04x",
+			     address);
+		goto out;
+	}
+	len = g_ascii_strtoull (values[1], NULL, 10);
+	if (len < 1) {
+		ret = FALSE;
+		g_set_error (error, 1, 0,
+			     "invalid length %" G_GSIZE_FORMAT,
+			     len);
+		goto out;
+	}
+	data = g_new0 (guint8, len);
+	ch_device_queue_read_sram (priv->device_queue,
+				   priv->device,
+				   (guint16) address,
+				   data,
+				   len);
+	ret = ch_device_queue_process (priv->device_queue,
+				       CH_DEVICE_QUEUE_PROCESS_FLAGS_NONE,
+				       NULL,
+				       error);
+	if (!ret)
+		goto out;
+
+	g_print ("Read:\n");
+	for (i = 0; i < len; i++)
 		g_print ("0x%04x = %02x\n", address + i, data[i]);
 out:
 	g_free (data);
@@ -2712,6 +3204,11 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Takes all color readings (to XYZ)"),
 		     ch_util_take_readings_xyz);
+	ch_util_add (priv->cmd_array,
+		     "take-reading-spectral",
+		     /* TRANSLATORS: command description */
+		     _("Takes a spectral reading (saving to SRAM)"),
+		     ch_util_take_reading_spectral);
 	ch_util_add (priv->cmd_array,
 		     "reset",
 		     /* TRANSLATORS: command description */
@@ -2802,11 +3299,56 @@ main (int argc, char *argv[])
 		     /* TRANSLATORS: command description */
 		     _("Uploads a remote profile"),
 		     ch_util_remote_profile_upload);
+	ch_util_add (priv->cmd_array,
+		     "ccmx-upload",
+		     /* TRANSLATORS: command description */
+		     _("Uploads a correction matrix"),
+		     ch_util_ccmx_upload);
 	ch_util_add (priv->cmd_array,
 		     "self-test",
 		     /* TRANSLATORS: command description */
 		     _("Does a quick self test on the device"),
 		     ch_util_self_test);
+	ch_util_add (priv->cmd_array,
+		     "get-measure-mode",
+		     /* TRANSLATORS: command description */
+		     _("Gets the sensor measurement mode"),
+		     ch_util_get_measure_mode);
+	ch_util_add (priv->cmd_array,
+		     "set-measure-mode",
+		     /* TRANSLATORS: command description */
+		     _("Sets the sensor measurement mode"),
+		     ch_util_set_measure_mode);
+	ch_util_add (priv->cmd_array,
+		     "sram-read",
+		     /* TRANSLATORS: command description */
+		     _("Read SRAM at a specified address"),
+		     ch_util_sram_read);
+	ch_util_add (priv->cmd_array,
+		     "sram-write",
+		     /* TRANSLATORS: command description */
+		     _("Write SRAM at a specified address"),
+		     ch_util_sram_write);
+	ch_util_add (priv->cmd_array,
+		     "get-temperature",
+		     /* TRANSLATORS: command description */
+		     _("Gets the sensor temperature"),
+		     ch_util_get_temperature);
+	ch_util_add (priv->cmd_array,
+		     "get-adc-vrefs",
+		     /* TRANSLATORS: command description */
+		     _("Gets the ADC Vref values"),
+		     ch_util_get_adc_vrefs);
+	ch_util_add (priv->cmd_array,
+		     "get-ccd-calibration",
+		     /* TRANSLATORS: command description */
+		     _("Gets the CCD calibration values"),
+		     ch_util_get_ccd_calibration);
+	ch_util_add (priv->cmd_array,
+		     "set-ccd-calibration",
+		     /* TRANSLATORS: command description */
+		     _("Sets the CCD calibration values"),
+		     ch_util_set_ccd_calibration);
 
 	/* sort by command name */
 	g_ptr_array_sort (priv->cmd_array,
